package fractsplinewavelets;

import imageware.ImageWare;
import imageware.Builder;

/**
 * <hr>
 * <p><b>Plugin of ImageJ:</b><br>
 * Fractional Spline Wavelets<br>
 * <a href="http://bigwww.epfl.ch/demo/fractsplines/">More info</a>
 *
 * <p><b>Reference:</b><br>
 * M. Unser, T. Blu, "Fractional Splines and Wavelets," SIAM Review, vol. 42, no. 1, pp. 43-67, March 2000.<br>
 *
 * <p><b>Contact:</b><br>
 * Daniel Sage <a href="mailto:daniel.sage@epfl.ch?subject=Fractional Spline Wavelet Plugin">daniel.sage@epfl.ch</a><br>
 *
 * <p><b>Author:</b><br>
 * Gil Gaillard, Dimitri Van De Ville, Daniel Sage
 *
 * <p><b>Organisation:</b><br>
 * Swiss Federal Institute of Technology Lausanne,
 * Biomedical Imaging Group,
 * CH-1015 Lausanne, Switzerland,
 * <a href="http://bigwww.epfl.ch">http://bigwww.epfl.ch</a><br>
 *
 * <p><b>Version:</b><br>
 * 24 January 2007<br>
 *
 * <p><b>Copyright</b><br>
 * Copyright  2002-2007, Ecole Polytechnique Federale de Lausanne (EPFL), Lausanne, Switzerland, (EPFL)<br>
 *
 * <hr>
 *
 * <p><b>How to compile your own processing function</b><br>
 * 1) Implement your algorithm in the function "DoProcessing3D" below. 
 * 2) Compile the modified java source file with a Java compiler, e.g.:
 *       javac -classpath ../data/data.jar:../../../ij.jar CoefProcessing.java
 *    The ImageAccess library (data.jar) can be found in the archive of this plug-in.
 *    You also need the jar-version (ij.jar) of ImageJ, downloadable from the ImageJ site.
 * 3) Start the plug-in, and select "Your own processing"  
 */

public class CoefProcessing {

	/**
	* Template for processing 3D wavelet coefficients.
	*
	* @param coef			a ImageWare object representing the wavelet coefficients
	* @param iterations	the number of iterations along each dimension
	*/
	static public void doProcessing3D(ImageWare coef, int[] iterations) {
		// get the iterations along each dimension
		int iterationX=iterations[0];
		int iterationY=iterations[1];
		int iterationZ=iterations[2];

		// get the size along each dimension
		int nx = coef.getSizeX();
		int ny = coef.getSizeY();
		int nz = coef.getSizeZ();
		
		int type = coef.getType();
		
		// auxilary matrix to facilitate the loop processing each 3D subband
	    boolean offsets[][] = {
	      /* LLL */ { false, false, false },
	      /* LLH */ { false, false, true  },
	      /* LHL */ { false, true,  false },
	      /* LHH */ { false, true,  true  },
	      /* HLL */ { true,  false, false },
	      /* HLH */ { true,  false, true  },
	      /* HHL */ { true,  true,  false },
	      /* HHH */ { true,  true,  true  }
	    };

		// data block to process subband
		ImageWare sub;

		// outer loop: continue processing up to maximal number of iterations	
		for ( int i=0; i<Math.max(iterationX,Math.max(iterationY,iterationZ)); i++) {

			// adapt dimensions of subband along a dimensions
			if (iterationX>i) nx = nx / 2;
			if (iterationY>i) ny = ny / 2;
			if (iterationZ>i) nz = nz / 2;
			
			// create empty block at size of subband
			sub = Builder.create(nx, ny, nz, type);
			
			// process all subbands except LLL (2^3-1)
			for ( int j=1; j<8; j++ ) {
			
				// do only try to access a highpass along a direction if it is available
				if ( !( (iterationX<=i && offsets[j][0]) ||
					    (iterationY<=i && offsets[j][1]) ||
					    (iterationZ<=i && offsets[j][2]) ) ) 
				{
					// read the subband in to the data block
					coef.getXYZ( (offsets[j][0]) ? nx : 0,
							  (offsets[j][1]) ? ny : 0,
							  (offsets[j][2]) ? nz : 0 , sub);
							
					// ...  
					// ...your processing here... 
					// ...
					
					// write out the processed subband
					coef.putXYZ( (offsets[j][0]) ? nx : 0,
							  (offsets[j][1]) ? ny : 0,
							  (offsets[j][2]) ? nz : 0 , sub);
				}
			}
		}

		// finally process the remaining lowpass band
		sub = Builder.create(nx,ny,nz,type);

		// ...
		// ...your processing here...
		// ...

		coef.putXYZ(0, 0, 0, sub);
	}

	/**
	* Apply a soft thresholding to all coefficients.
	*
	* All the image values lower than '-threshold' and greater than 'threshold'
	* are set to 0. The remaining positive values are reduced by 'threshold';
	* the remaining negative values are augmented by 'threshold'.
	*
	* @param coef			a ImageWare object representing the wavelet coefficients
	* @param threshold      a value for the threshold
	*/
	static public void doSoftThreshold3D(ImageWare coef, double threshold) {
		int nx = coef.getSizeX();
		int ny = coef.getSizeY();
		int nz = coef.getSizeZ();
		double pixel=0.0;
		
		double[] row = new double[nx];
		for (int z=0; z<nz; z++)
	  	  for (int y=0; y<ny; y++) {
			coef.getX(0, y, z, row);
			for (int x=0; x<nx; x++) {
				if (row[x] < -threshold)
					row[x] = row[x] + threshold;
				else if (row[x] > threshold)
					row[x] = row[x] - threshold;
				else
					row[x] = 0.0;
			}
			coef.putX(0, y, z, row);
		  }
	}

	/**
	* Apply a hard thresholding to all coefficients.
	*
	* All the image values lower than '-threshold' and greater than 'threshold'
	* are set to 0.
	*
	* @param coef			a ImageWare representing the wavelet coefficients
	* @param threshold      a value for the threshold
	*/
	static public void doHardThreshold3D(ImageWare coef, double threshold) {
		int nx = coef.getSizeX();
		int ny = coef.getSizeY();
		int nz = coef.getSizeZ();
		
		double pixel=0.0;
		
		double[] row = new double[nx];
		for (int z=0; z<nz; z++)
		  for (int y=0; y<ny; y++) {
			coef.getX(0, y, z, row);
			for (int x=0; x<nx; x++) {
				if ((row[x] >= -threshold) && (row[x] <= threshold)) {
					row[x] = 0.0;
				}
			}
			coef.putX(0, y, z, row);
		  }
	}

	/**
	* Compute the number of non zero pixels.
	*
	* @param coef			a ImageWare object representing the wavelet coefficients
	* @param factor			the enhancement factor	
	*/
	static public void doEnhancement3D(ImageWare coef, double factor, int iterations[]) {
		int nx0 = coef.getSizeX()/(int)Math.pow((double)2,(double)(iterations[0]));
		int ny0 = coef.getSizeY()/(int)Math.pow((double)2,(double)(iterations[1]));
		int nz0 = coef.getSizeZ()/(int)Math.pow((double)2,(double)(iterations[2]));
		
		ImageWare sub = Builder.create(nx0, ny0, nz0, coef.getType());
		ImageWare coefmultiply = null;
		coef.getXYZ(0,0,0, sub);
		sub.multiply(factor);
		coef.putXYZ(0,0,0,sub);
	}

	/**
	* Apply a rescaling.
	*
	* Stretches the contrast inside subbands
	* so that the gray levels are in the range : the positivy values
	* are stretched to center-max and the negative values are stretched 
	* to min-center and a contrast stretching is applied on the low band
	*
	* @param coef			a ImageWare object representing the wavelet coefficients
	* @param iterations    	number of iterations
	*/
	static final public void doRescale3D(ImageWare coef, int[] iterations) {
		int iterationX=iterations[0];
		int iterationY=iterations[1];
		int iterationZ=iterations[2];

		int nx = coef.getSizeX();
		int ny = coef.getSizeY();
		int nz = coef.getSizeZ();
		
		final int type = coef.getType();
		
	    final boolean offsets[][] = {
	      /* LLL */ { false, false, false },
	      /* LLH */ { false, false, true  },
	      /* LHL */ { false, true,  false },
	      /* LHH */ { false, true,  true  },
	      /* HLL */ { true,  false, false },
	      /* HLH */ { true,  false, true  },
	      /* HHL */ { true,  true,  false },
	      /* HHH */ { true,  true,  true  }
	    };

		ImageWare sub;
		double subd[][][];
		
		for ( int i=0; i<Math.max(iterationX,Math.max(iterationY,iterationZ)); i++) {
			if (iterationX>i) nx = nx / 2;
			if (iterationY>i) ny = ny / 2;
			if (iterationZ>i) nz = nz / 2;
			
			sub = Builder.create(nx, ny, nz, type);
			
			// all subbands except LLL
			for ( int j=1; j<8; j++ ) {
				if ( !( (iterationX<=i && offsets[j][0]) ||
					    (iterationY<=i && offsets[j][1]) ||
					    (iterationZ<=i && offsets[j][2]) ) ) {
					coef.getXYZ( (offsets[j][0]) ? nx : 0,
							  (offsets[j][1]) ? ny : 0,
							  (offsets[j][2]) ? nz : 0 , sub); 
					normalizeContrastCentered(sub);
					coef.putXYZ( (offsets[j][0]) ? nx : 0,
							  (offsets[j][1]) ? ny : 0,
							  (offsets[j][2]) ? nz : 0 , sub);
				}
			}
		}

		// lowest LLL
		sub = Builder.create(nx,ny,nz,type);
		coef.getXYZ(0, 0, 0, sub);
		sub.rescale();
		coef.putXYZ(0, 0, 0, sub);
	}

	/**
	* Stretches the contrast inside an image so that the gray levels are
	* in the range : the positivy values are stretched to center-max 
	* and the negative values are stretched to min-center.
	*/
	private static void normalizeContrastCentered(ImageWare image) {
		double maxGoal = 255.0;
		double minGoal = 0.0;
		double center = (maxGoal+minGoal)/2.0;
		
		// Get the min and max
		double minImage = image.getMinimum();
		double maxImage = image.getMaximum();
		
		// Compute the parameter to rescale the gray levels
		double a;
		if (minImage-maxImage == 0) {
			a = 1.0;
			minImage = (maxGoal-minGoal)/2.0;
		}
		else {
			if ( Math.abs(maxImage) > Math.abs(minImage) )
				a = (maxGoal-center) / Math.abs(maxImage);
			else
				a = (center-minGoal) / Math.abs(minImage);
		}
		
		final int nx=image.getSizeX();
		final int ny=image.getSizeY();
		final int nz=image.getSizeZ(); 
		
		double row[]=new double[nx];
		
		for (int z=0; z<nz; z++)
		  for (int y=0; y<ny; y++) {
			image.getX(0, y, z, row);
			for (int x=0; x<nx; x++) {
				row[x] *= a;
				row[x] += center;
			}
			image.putX(0, y, z, row);
		  }
					
	}

	/**
	* Apply a rescaling.
	*
	* Stretches the contrast inside subbands
	* so that the gray levels are in the range : the positivy values
	* are stretched to center-max and the negative values are stretched 
	* to min-center and a contrast stretching is applied on the low band
	*
	* @param coef			a ImageWare object representing the wavelet coefficients
	* @param iterations    	number of iterations
	*/
	static final public void doRescale3DNonCenter(ImageWare coef, int[] iterations)  {
		int iterationX=iterations[0];
		int iterationY=iterations[1];
		int iterationZ=iterations[2];

		int nx = coef.getSizeX();
		int ny = coef.getSizeY();
		int nz = coef.getSizeZ();
		
		final int type = coef.getType();
		
	    final boolean offsets[][] = {
	      /* LLL */ { false, false, false },
	      /* LLH */ { false, false, true  },
	      /* LHL */ { false, true,  false },
	      /* LHH */ { false, true,  true  },
	      /* HLL */ { true,  false, false },
	      /* HLH */ { true,  false, true  },
	      /* HHL */ { true,  true,  false },
	      /* HHH */ { true,  true,  true  }
	    };

		ImageWare sub;
		double subd[][][];
		
		for ( int i=0; i<Math.max(iterationX,Math.max(iterationY,iterationZ)); i++) {
			if (iterationX>i) nx = nx / 2;
			if (iterationY>i) ny = ny / 2;
			if (iterationZ>i) nz = nz / 2;
			
			sub = Builder.create(nx, ny, nz, type);
			
			// all subbands except LLL
			for ( int j=1; j<8; j++ ) {
				if ( !( (iterationX<=i && offsets[j][0]) ||
					    (iterationY<=i && offsets[j][1]) ||
					    (iterationZ<=i && offsets[j][2]) ) ) {
					coef.getXYZ( (offsets[j][0]) ? nx : 0,
							  (offsets[j][1]) ? ny : 0,
							  (offsets[j][2]) ? nz : 0 , sub); 
					normalizeContrastNonCentered(sub);
					coef.putXYZ( (offsets[j][0]) ? nx : 0,
							  (offsets[j][1]) ? ny : 0,
							  (offsets[j][2]) ? nz : 0 , sub);
				}
			}
		}

		// lowest LLL
		sub = Builder.create(nx,ny,nz,type);
		coef.getXYZ(0, 0, 0, sub);
		sub.rescale();
		coef.putXYZ(0, 0, 0, sub);
	}

	/**
	* Stretches the contrast inside an image so that the gray levels are
	* in the range thevalues are stretched to min-max. 
	*/
	private static void normalizeContrastNonCentered(ImageWare image) {
		double minGoal = 0.0;
		double maxGoal = 255.0;
		
		// Get the min and max
		double minImage = image.getMinimum();
		double maxImage = image.getMaximum();
		
		// Compute the parameter to rescale the gray levels
		double a;
		if ( minImage-maxImage == 0) {
			a = 1.0;
			minImage = (maxGoal-minGoal)/2.0;
		}
		else
			a = (maxGoal-minGoal) / (maxImage-minImage);

		final int nx=image.getSizeX();
		final int ny=image.getSizeY();
		final int nz=image.getSizeZ(); 
		
		double row[]=new double[nx];
		
		for (int z=0; z<nz; z++)
		  for (int y=0; y<ny; y++) {
			image.getX(0, y, z, row);
			for (int x=0; x<nx; x++) {
				row[x] = row[x] - minGoal;
			}
			image.putX(0, y, z, row);
		  }
					
	}

}